//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"
//---------------------------------------------------------------------------
int parseNumber(char *str);
int readFile(char *fileName, BYTE *buff, int maxLen);
int writeFile(char *fileName, BYTE *buff, int len, BOOL append);
int printError(int err, void *par);
void version(void);
void usage(void);

//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  char *fnIn = NULL, *fnOut = NULL, fn[256], n[11], *pc;
  BOOL vb = FALSE, append = FALSE, pmd = FALSE, i8080 = FALSE, vram = FALSE;
  BOOL z80 = FALSE, zx = FALSE, bas = FALSE, d80 = FALSE, hp = FALSE;
  int colors[] = { 0, 0, 7 }, clearAddr = -1, runAddr = -1;
  int targetAddr = -1, unpackRtnAddr = -1, jumpAddr = -1;
  BYTE buff1[65536], buff2[65536], *pb;
  int i, err, mode, lenIn, lenOut[4], lenRt[8], lenTot;
  int useModes = (PACK_MODE_1 | PACK_MODE_2 | PACK_MODE_3 | PACK_MODE_4);

  // spracovanie a kontrola vstupnych parametrov
  err = 0;
  do {
    for (i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-h") == 0) {
        err = ERR_PRINT_USAGE;
        break;
      }
      else if (strcmp(argv[i], "-v") == 0) {
        err = ERR_PRINT_VERSION;
        break;
      }
      else if (strcmp(argv[i], "-vb") == 0) {
        vb = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-z80") == 0) {
        z80 = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-i8080") == 0) {
        i8080 = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-ptp") == 0) {
        pmd = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-vram") == 0) {
        vram = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-tap") == 0) {
        zx = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-b") == 0) {
        bas = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-d80") == 0) {
        d80 = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-hp") == 0) {
        hp = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-append") == 0) {
        append = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-m") == 0) {
        useModes = 0;
        if ((i + 1) < argc) {
          pc = argv[++i];
          while (*pc) {
            if (*pc < '1' || *pc > '4')
              break;
            if (useModes & (1 << (*pc - '1')))
              break;
            useModes |= (1 << (*pc - '1'));
            pc++;
          }
        }
        if (useModes == 0 || *pc != 0) {
          err = printError(ERR_BAD_PACK_MODE, argv[i]);
          break;
        }
        continue;
      }
      else if (strcmp(argv[i], "-cb") == 0) {
        if (++i == argc)
          colors[0] = -1;
        else
          colors[0] = parseNumber(argv[i]);
        if (colors[0] < 0 || colors[0] > 7) {
          err = printError(ERR_BAD_COLOR_VALUE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-cp") == 0) {
        if (++i == argc)
          colors[1] = -1;
        else
          colors[1] = parseNumber(argv[i]);
        if (colors[1] < 0 || colors[1] > 7) {
          err = printError(ERR_BAD_COLOR_VALUE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-ci") == 0) {
        if (++i == argc)
          colors[2] = -1;
        else
          colors[2] = parseNumber(argv[i]);
        if (colors[2] < 0 || colors[2] > 7) {
          err = printError(ERR_BAD_COLOR_VALUE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-a") == 0) {
        if (++i == argc)
          targetAddr = -1;
        else
          targetAddr = parseNumber(argv[i]);
        if (targetAddr < 0 || targetAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-ra") == 0) {
        if (++i == argc)
          unpackRtnAddr = -1;
        else
          unpackRtnAddr = parseNumber(argv[i]);
        if (unpackRtnAddr < 0 || unpackRtnAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-ja") == 0) {
        if (++i == argc)
          jumpAddr = -1;
        else
          jumpAddr = parseNumber(argv[i]);
        if (jumpAddr < 0 || jumpAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-c") == 0) {
        if (++i == argc)
          clearAddr = -1;
        else
          clearAddr = parseNumber(argv[i]);
        if (clearAddr < 0 || clearAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-r") == 0) {
        if (++i == argc)
          runAddr = -1;
        else
          runAddr = parseNumber(argv[i]);
        if (runAddr < 0 || runAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-o") == 0) {
        if (++i == argc || fnOut != NULL) {
          err = printError(ERR_OUTPUT_FILE_MORE, NULL);
          break;
        }
        fnOut = argv[i];
        continue;
      }
      else {
        if (fnIn != NULL) {
          err = printError(ERR_INPUT_FILE_MORE, NULL);
          break;
        }
        fnIn = argv[i];
        continue;
      }
    }
    if (err != 0)
      break;
    if ((zx && (pmd || i8080)) || (z80 && i8080) || (vram && (!i8080 || z80 || zx))
        || (pmd && (z80 || bas || hp || d80 || runAddr >= 0 || clearAddr >= 0))
        || (!zx && (bas || hp || d80 || runAddr >= 0 || clearAddr >= 0))) {
      err = printError(ERR_BAD_OPTIONS, NULL);
      break;
    }
    if (fnIn == NULL) {
      err = printError(ERR_NO_INPUT_FILE, NULL);
      break;
    }
    if (fnOut == NULL) {
      strcpy(fn, fnIn);
      pc = strrchr(fn, '.');
      if (pc != NULL)
        *pc = 0;
      if (zx)
        strcat(fn, ".tap");
      else if (pmd)
        strcat(fn, ".ptp");
      else
        strcat(fn, ".pck");
      fnOut = fn;
      if (strcmp(fnOut, fnIn) == 0)
        strcat(fnOut, ".pck");
    }

    // default hodnoty
    if (z80 || i8080) {
      if (targetAddr < 0)
        targetAddr = ((z80) ? 32768 : 0);
      if (unpackRtnAddr < 0)
        unpackRtnAddr = ((z80) ? 23456 : 0x7F00);
      if ((i8080 && vram && unpackRtnAddr >= 0xC030 && (unpackRtnAddr & 0x3F) != 0x30)
          || (i8080 && vram && (unpackRtnAddr < 0xC030 || unpackRtnAddr > 0xFDF0))) {
        err = printError(ERR_BAD_UNPACK_ADDR, (int *)unpackRtnAddr);
        break;
      }
    }
    if (zx && runAddr < 0)
      runAddr = 32768;
    if (zx && clearAddr < 0)
      clearAddr = 24575;

    // ziskanie dlzok rozpakovacich rutin
    if (z80)
      getRoutinesLengthZ80(lenRt);
    else if (i8080)
      getRoutinesLengthI8080(lenRt, vram);
    else
      memset(lenRt, 0, sizeof(lenRt));

    // precitanie vstupneho suboru
    if (vb)
      printf("Reading input file '%s'\n", fnIn);
    lenIn = readFile(fnIn, buff1, MAX_FILE_LENGTH);
    if (lenIn < 0) {
      err = lenIn;
      break;
    }
    if (vb)
      printf("Source data length: %d bytes\n", lenIn);

    // spakovanie dat zvolenymi modmi
    for (i = 0; i < 4; i++) {
      mode = useModes & (1 << i);
      if (mode) {
        if (vb)
          printf("Packing in mode: %d\n", i + 1);
        memcpy(buff2, buff1, sizeof(buff2));
        lenOut[i] = pack(buff2, lenIn, mode);

        if (vb) {
          if (lenOut[i] < 0)
            printf("  Packing ineffective!\n");
          else {
            printf("  Packed data length: %d bytes", lenOut[i]);
            if (z80 || i8080) {
              lenTot = lenRt[i * 2] + lenRt[i * 2 + 1];
              printf("\n  Unpack routine length: %d bytes, without mover: %d bytes", lenTot, lenRt[i * 2 + 1]);
              if (vram)
                printf(" (%d VRAM blocks)", (lenRt[i * 2 + 1] + 15) / 16);
              lenTot += lenOut[i];
              printf("\n  Total length: %d bytes", lenTot);
            }
            else
              lenTot = lenOut[i];
            printf(" (%.2f%%)\n", (lenTot * 100.0) / lenIn);
            lenOut[i] = lenTot;
          }
        }
        else if (lenOut[i] > 0 && (z80 || i8080)) {
          lenTot = lenRt[i * 2] + lenRt[i * 2 + 1] + lenOut[i];
          lenOut[i] = lenTot;
        }
      }
      else
        lenOut[i] = -1;
    }

    // najdenie najlepsieho vysledku
    lenTot = lenIn;
    mode = -1;
    for (i = 0; i < 4; i++) {
      if (lenOut[i] > 0 && lenOut[i] < lenTot) {
        lenTot = lenOut[i];
        mode = i;
      }
    }
    if (mode < 0) {
      err = printError(ERR_INEFFECTIVE, NULL);
      break;
    }

    // konecne pakovanie najlepsim modom
    memcpy(buff2, buff1, sizeof(buff2));
    lenOut[0] = pack(buff2, lenIn, 1 << mode);
    lenTot = lenOut[0];
    if (z80 || i8080)
      lenTot += (lenRt[mode * 2] + lenRt[mode * 2 + 1]);
    printf("Packed in mode: %d\n", mode + 1);
    printf("Original length: %d bytes\n", lenIn);
    printf("New length: %d bytes (%.2f%%)\n", lenTot, (lenTot * 100.0) / lenIn);

    pb = buff1;
    if (zx || pmd) {
      if (vb)
        printf("Constructing %s file\n", (zx) ? "TAP" : "PTP");

      pc = strrchr(fnIn, '/');
      if (pc == NULL)
        pc = strrchr(fnIn, '\\');
      if (pc == NULL)
        pc = fnIn;
      else
        pc++;
      i = (zx) ? 10 : 8;
      strncpy(n, pc, i);
      n[i] = 0;
      pc = strrchr(n, '.');
      if (pc != NULL)
        *pc = 0;

      if (zx) {
        // ZX Spectrum TAP
        if (bas)
          constructBasicZX(&pb, colors, clearAddr, hp, n, d80, runAddr);
        constructCodeZX(&pb, buff2, lenOut[0], lenIn, (z80) ? (1 << mode) : 0,
                        n, targetAddr, unpackRtnAddr, jumpAddr);
        lenTot = pb - buff1;
      }
      else { // if (pmd)
        // PMD 85 PTP
        constructCodePMD(&pb, buff2, lenOut[0], lenIn, n,
                         (i8080) ? (1 << mode) : 0, vram,
                         targetAddr, unpackRtnAddr, jumpAddr);
        lenTot = pb - buff1;
      }

    }
    else if (z80) {
      constructBinZ80(&pb, buff2, lenOut[0], lenIn, 1 << mode,
                    targetAddr, unpackRtnAddr, jumpAddr);
      lenTot = pb - buff1;
    }
    else if (i8080) {
      constructBinI8080(&pb, buff2, lenOut[0], lenIn, 1 << mode, vram,
                       targetAddr, unpackRtnAddr, jumpAddr);
      lenTot = pb - buff1;
    }
    else
      memcpy(buff1, buff2, sizeof(buff1));

    // zapis vystupneho suboru
    if (vb)
      printf("Writing to the output file '%s'\n", fnOut);
    err = writeFile(fnOut, buff1, lenTot, append);

  } while (FALSE);

  if (err == ERR_PRINT_USAGE) {
    usage();
    err = 0;
  }
  if (err == ERR_PRINT_VERSION) {
    version();
    err = 0;
  }

  return -err;
}
//---------------------------------------------------------------------------
int parseNumber(char *str)
{
  char *p;
  int val = strtol(str, &p, 0);
  if (val < 0 || (val == 0 && *p != 0))
    val = -1;
  return val;
}
//---------------------------------------------------------------------------
int readFile(char *fileName, BYTE *buff, int maxLen)
{
  int err = 0, len;
  FILE *f;

  do {
    f = fopen(fileName, "rb");
    if (f == NULL) {
      err = printError(ERR_OPEN_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_END) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
    len = ftell(f);
    if (len < 0 || len > maxLen) {
      err = printError(ERR_LENGTH_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_SET) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
    if (fread(buff, 1, len, f) != (unsigned int)len) {
      err = printError(ERR_READ_SRC_FILE, fileName);
      break;
    }
  } while (FALSE);

  fclose(f);

  if  (err < 0)
    return err;

  return len;
}
//---------------------------------------------------------------------------
int writeFile(char *fileName, BYTE *buff, int len, BOOL append)
{
  int err = 0;
  FILE *f;

  do {
    if (append)
      f = fopen(fileName, "ab");
    else
      f = fopen(fileName, "wb");
    if (f == NULL) {
      err = printError(ERR_OPEN_DEST_FILE, fileName);
      break;
    }
    if (!append && fseek(f, 0L, SEEK_SET) != 0) {
      err = printError(ERR_ACCESS_DEST_FILE, fileName);
      break;
    }
    if (fwrite(buff, 1, len, f) != (unsigned int)len) {
      err = printError(ERR_WRITE_DEST_FILE, fileName);
      break;
    }
  } while (FALSE);

  fclose(f);

  return err;
}
//---------------------------------------------------------------------------
int printError(int err, void *par)
{
  switch (err) {
    case ERR_BAD_PACK_MODE :
      fprintf(stderr, "Bad packing mode specified: '%s'!", (char *)par);
      break;

    case ERR_BAD_OPTIONS :
      fprintf(stderr, "Bad combination of an options!");
      break;

    case ERR_OPEN_SRC_FILE :
      fprintf(stderr, "Can't open input file '%s'!", (char *)par);
      break;

    case ERR_ACCESS_SRC_FILE :
      fprintf(stderr, "Can't access input file '%s'!", (char *)par);
      break;

    case ERR_LENGTH_SRC_FILE :
      fprintf(stderr, "Can't retrieve length of the input file '%s'!", (char *)par);
      break;

    case ERR_READ_SRC_FILE :
      fprintf(stderr, "Can't read from the input file '%s'!", (char *)par);
      break;

    case ERR_OPEN_DEST_FILE :
      fprintf(stderr, "Can't open output file '%s'!", (char *)par);
      break;

    case ERR_ACCESS_DEST_FILE :
      fprintf(stderr, "Can't access output file '%s'!", (char *)par);
      break;

    case ERR_WRITE_DEST_FILE :
      fprintf(stderr, "Can't write to the output file '%s'!", (char *)par);
      break;

    case ERR_BAD_COLOR_VALUE :
      fprintf(stderr, "After '%s' must follow color in range 0 to 7!", (char *)par);
      break;

    case ERR_BAD_ADDRESS :
      fprintf(stderr, "After '%s' must follow address in range 0 to 65535!", (char *)par);
      break;

    case ERR_BAD_UNPACK_ADDR :
      fprintf(stderr, "Bad address of the unpack routine on PMD 85: 0x%04X\n", (int)par);
      fprintf(stderr, "Address of the VRAM located unpack routine must satisfy following condition:\n");
      fprintf(stderr, "  addr >= 0xC030 AND addr <= 0xFDF0 AND (addr & 0xC03F) = 0xC030");
      break;

    case ERR_INEFFECTIVE :
      fprintf(stderr, "Packing ineffective!");
      break;

    case ERR_NO_INPUT_FILE :
      fprintf(stderr, "Input file not specified!\nTry 'quido -h' for help.");
      break;

    case ERR_INPUT_FILE_MORE :
      fprintf(stderr, "Only one input file can be specified!");
      break;

    case ERR_OUTPUT_FILE_MORE :
      fprintf(stderr, "Only one output file can be specified!");
      break;
  }

  fprintf(stderr, "\n\n");

  return err;
}
//---------------------------------------------------------------------------
void version(void)
{
	printf("%s v%d.%d - %s\n",
         PROG_NAME, PROG_VERSION / 16, PROG_VERSION % 16, PROG_AUTHOR);
}
//---------------------------------------------------------------------------
void pressEnter(void)
{
  printf("\nPress ENTER ...");
  while (getchar() != '\n')
    ;
}
//---------------------------------------------------------------------------
void usage(void)
{
  version();

	printf("Usage: %s [options] infile.ext\n", PROG_NAME);
  printf("\nCommon options:\n");
  printf("  -o outfile.ext  output file name\n");
  printf("  -append\t  append at the end of the output file\n");
  printf("  -m 1234\t  pack only with given modes [all modes are trying]\n");
  printf("  -z80\t\t  insert Z80 unpack routine\n");
  printf("  -i8080\t  insert i8080 unpack routine\n");
  printf("  -a address\t  start address of binary file [32768 (z80) / 0 (i8080)]\n");
  printf("  -ra address\t  unpack routine address [23456 (z80) / 32512 (i8080)]\n");
  printf("  -ja address\t  jump address after unpack [ret]\n");
  printf("  -vb\t\t  verbose\n");
  printf("  -h\t\t  usage information\n");
  printf("  -v\t\t  version info\n");

  printf("\nOptions for PTP file (PMD 85):\n");
  printf("  -ptp\t\t  create PTP file\n");
  printf("  -vram\t\t  locate unpack routine to the VideoRAM\n");

  pressEnter();

  printf("\nOptions for TAP file (ZX Spectrum):\n");
  printf("  -tap\t\t  create TAP file\n");
  printf("  -b\t\t  include BASIC loader\n");
  printf("  -c address\t  CLEAR address in BASIC loader [24575]\n");
  printf("  -r address\t  address where to start bin. file for BASIC loader [32768]\n");
  printf("  -cb colour\t  border color set by loader [0]\n");
  printf("  -cp colour\t  paper color set by loader [0]\n");
  printf("  -ci colour\t  ink color set by loader [7]\n");
  printf("  -d80\t\t  create D80 syntax loader\n");
  printf("  -hp\t\t  include POKE command for disabling tape headers\n");

  printf("\nAll numeric parameters must be nonnegative integers and can be in three forms:\n");
  printf("  decimal (e.g. 12345)\n");
  printf("  octal (e.g. 01234)\n");
  printf("  hexadecimal (e.g. 0x1234)\n\n");

  return;
}
//---------------------------------------------------------------------------

